"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.printOutboxPollingListenerEnvVariables = exports.printInboxPollingListenerEnvVariables = exports.getOutboxPollingListenerSettings = exports.getInboxPollingListenerSettings = void 0;
const env_settings_1 = require("../common/env-settings");
const listener_config_1 = require("../common/listener-config");
const basicSettingsMap = [
    {
        constantName: 'NEXT_MESSAGES_FUNCTION_SCHEMA',
        default: 'public',
        func: env_settings_1.getEnvVariableString,
    },
    {
        constantName: 'NEXT_MESSAGES_BATCH_SIZE',
        default: 5,
        func: env_settings_1.getEnvVariableNumber,
    },
    {
        constantName: 'NEXT_MESSAGES_LOCK_IN_MS',
        default: 5000,
        func: env_settings_1.getEnvVariableNumber,
    },
    {
        constantName: 'NEXT_MESSAGES_POLLING_INTERVAL_IN_MS',
        default: 500,
        func: env_settings_1.getEnvVariableNumber,
    },
];
const inboxSettingsMap = [
    {
        constantName: 'NEXT_MESSAGES_FUNCTION_NAME',
        default: 'next_inbox_messages',
        func: env_settings_1.getEnvVariableString,
    },
];
const outboxSettingsMap = [
    {
        constantName: 'NEXT_MESSAGES_FUNCTION_NAME',
        default: 'next_outbox_messages',
        func: env_settings_1.getEnvVariableString,
    },
];
/**
 * Loads the environment variables into the polling listener settings object. It
 * supports reading an inbox specific setting or a general one.
 * Please use the `printInboxPollingListenerEnvVariables` functions to get a
 * list of all the inbox relevant settings for the polling listener.
 * @example
 * TRX_DB_SCHEMA=trx_schema
 * TRX_INBOX_DB_TABLE=inbox_table
 * TRX_INBOX_NEXT_MESSAGES_FUNCTION_SCHEMA=next_inbox_messages
 * @param env The process.env variable or a custom object.
 * @returns The polling listener settings object with filled with the ENV variables
 */
const getInboxPollingListenerSettings = (env = process.env) => {
    const inboxSettings = (0, listener_config_1.getInboxListenerSettings)(env);
    const pollSettings = (0, env_settings_1.getConfigSettings)([...basicSettingsMap, ...inboxSettingsMap], listener_config_1.inboxEnvPrefix, listener_config_1.fallbackEnvPrefix, env);
    return Object.assign(Object.assign({}, inboxSettings), pollSettings);
};
exports.getInboxPollingListenerSettings = getInboxPollingListenerSettings;
/**
 * Loads the environment variables into the polling listener settings object. It
 * supports reading an outbox specific setting or a general one.
 * Please use the `printOutboxPollingListenerEnvVariables` functions to get a
 * list of all the outbox relevant settings for the polling listener.
 * @example
 * TRX_DB_SCHEMA=trx_schema
 * TRX_OUTBOX_DB_TABLE=outbox_table
 * TRX_OUTBOX_NEXT_MESSAGES_FUNCTION_SCHEMA=next_outbox_messages
 * @param env The process.env variable or a custom object.
 * @returns The polling listener settings object with filled with the ENV variables
 */
const getOutboxPollingListenerSettings = (env = process.env) => {
    const outboxSettings = (0, listener_config_1.getOutboxListenerSettings)(env);
    const pollSettings = (0, env_settings_1.getConfigSettings)([...basicSettingsMap, ...outboxSettingsMap], listener_config_1.outboxEnvPrefix, listener_config_1.fallbackEnvPrefix, env);
    return Object.assign(Object.assign({}, outboxSettings), pollSettings);
};
exports.getOutboxPollingListenerSettings = getOutboxPollingListenerSettings;
/**
 * Shows the available env variables and their default values for the inbox
 * listener with the polling approach.
 * @param map A mapping of all the env variables to config settings.
 * @param envPrefix The prefix for the env variables to check first (e.g. "TRX_INBOX_" or "TRX_").
 * @param envPrefixFallback The fallback prefix if the other is not found. Useful for defining settings that should be used for both outbox and inbox.
 * @returns
 */
const printInboxPollingListenerEnvVariables = () => (0, listener_config_1.printInboxListenerEnvVariables)() +
    (0, env_settings_1.printConfigSettings)([...basicSettingsMap, ...inboxSettingsMap], listener_config_1.inboxEnvPrefix, listener_config_1.fallbackEnvPrefix);
exports.printInboxPollingListenerEnvVariables = printInboxPollingListenerEnvVariables;
/**
 * Shows the available env variables and their default values for the outbox
 * listener with the polling approach.
 * @param map A mapping of all the env variables to config settings.
 * @param envPrefix The prefix for the env variables to check first (e.g. "TRX_OUTBOX_" or "TRX_").
 * @param envPrefixFallback The fallback prefix if the other is not found. Useful for defining settings that should be used for both outbox and inbox.
 * @returns
 */
const printOutboxPollingListenerEnvVariables = () => (0, listener_config_1.printOutboxListenerEnvVariables)() +
    (0, env_settings_1.printConfigSettings)([...basicSettingsMap, ...outboxSettingsMap], listener_config_1.inboxEnvPrefix, listener_config_1.fallbackEnvPrefix);
exports.printOutboxPollingListenerEnvVariables = printOutboxPollingListenerEnvVariables;
//# sourceMappingURL=config.js.map